-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("achievementTracker", "AceEvent-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LATr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-AchievementTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local LibQTip = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local CreateFont, CreateFrame = CreateFont, CreateFrame
local type, unpack, format, pairs = type, unpack, string.format, pairs
local GetAchievementInfo, GetAchievementLink = GetAchievementInfo, GetAchievementLink
local GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements = GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements
local GetAchievementNumCriteria, GetAchievementCriteriaInfo = GetAchievementNumCriteria, GetAchievementCriteriaInfo
local UnitAffectingCombat = UnitAffectingCombat

-----------------
--- Variables ---
-----------------
module.modName = LATr["Achievement Tracker"]
local prdb, chdb, options
local numTrackedAchievement = 0
local category, achievements = {}, {}
local achievementTracker, trackerAnchor
local anchorPosition = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"}
local outlines = {[""] = LATr["None"], ["OUTLINE"] = LATr["Outline"], ["THICKOUTLINE"] = LATr["Thick Outline"]}
local cATrFont = CreateFont("cATrFont")
local scATrFont = CreateFont("scATrFont")
local aATrFont = CreateFont("aATrFont")
local acATrFont = CreateFont("acATrFont")
local COLOR_COPPER = "eda55f"
local COLOR_SILVER = "c7c7cf"
local COLOR_GOLD = "ffd700"

----------------------
-- Table recycling ---
----------------------
local tablePool = {}
setmetatable(tablePool, {__mode = "kv"})

local function new()
	local t = next(tablePool) or {}
	tablePool[t] = nil
	return t
end

local function del(t)
	if type(t) == "table" then
		for k, v in pairs(t) do
			if type(v) == "table" then
				v = del(v)
			end
			t[k] = nil
		end
		t[true] = true
		t[true] = nil
		setmetatable(t, nil)
		tablePool[t] = true
	end
	return nil
end

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(LATr["Achievement Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetClampedToScreen(true)
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end)

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedCriteria = true,
		categoryFont = LSM.DefaultMedia.font,
		categoryFontSize = 12,
		categoryFontOutline = "",
		categoryCollapseButtonSize = 24,
		subCategoryFont = LSM.DefaultMedia.font,
		subCategoryFontSize = 12,
		subCategoryFontOutline = "",
		subCategoryCollapseButtonSize = 24,
		achievementFont = LSM.DefaultMedia.font,
		achievementFontSize = 10,
		achievementFontOutline = "",
		achievementImageSize = 24,
		achievementCriteriaFont = LSM.DefaultMedia.font,
		achievementCriteriaFontSize = 8,
		achievementCriteriaFontOutline = "",
		numBooleanCirteriasColumns = 4,
		maxWidth = 300,
		subCategoryCellPadding = 5,
		achievementCellPadding = 10,
		criteriasCellPadding = 15,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, catFontSetter, subCatFontSetter, achFontSetter, achCritFontSetter, anchorSetter, combatSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function catFontSetter(info, value)
		prdb[info[#info]] = value
		cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function subCatFontSetter(info, value)
		prdb[info[#info]] = value
		scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function achFontSetter(info, value)
		prdb[info[#info]] = value
		aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function achCritFontSetter(info, value)
		prdb[info[#info]] = value
		acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function anchorSetter(info, value)
		prdb[info[#info]] = value
		if value then
			anchor:Show()
		else
			anchor:Hide()
		end
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
	function combatSetter(info, value)
		prdb[info[#info]] = value
		if value then
			module:RegisterEvent("PLAYER_REGEN_ENABLED", "updateAchievementsTracker")
			module:RegisterEvent("PLAYER_REGEN_DISABLED", "hideAchievementsTracker")
		else
			module:UnregisterEvent("PLAYER_REGEN_ENABLED")
			module:UnregisterEvent("PLAYER_REGEN_DISABLED")
		end
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				showAnchor = {
					name = LATr["Show achievement tracker anchor"],
					type = "toggle",
					set = anchorSetter,
					width = "full",
					order = 6,
				},
				anchorPosition = {
					name = LATr["Anchor position"],
					type = "select",
					values = anchorPosition,
					width = "full",
					order = 7,
				},
				toggleTracker = {
					name = LATr["Display tracker"],
					type = "toggle",
					width = "full",
					order = 9,
				},
				hideTrackerInCombat = {
					name = LATr["Hide achievement tracker in combat"],
					type = "toggle",
					set = combatSetter,
					width = "full",
					order = 10,
				},
				hideCompletedCriteria = {
					name = LATr["Hide completed achievement criteria"],
					type = "toggle",
					width = "full",
					order = 11,
				},
				categoryFont = {
					name = LATr["Achievement category font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = catFontSetter,
					width = "full",
					order = 12,
				},
				categoryFontSize = {
					name = LATr["Achievement category font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = catFontSetter,
					width = "full",
					order = 13,
				},
				categoryFontOutline = {
					name = LATr["Achievement category font outline"],
					type = "select",
					values = outlines,
					set = catFontSetter,
					width = "full",
					order = 14,
				},
				categoryCollapseButtonSize = {
					name = LATr["Category collapse Button size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 15,
				},
				subCategoryFont = {
					name = LATr["Achievement sub category font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = subCatFontSetter,
					width = "full",
					order = 16,
				},
				subCategoryFontSize = {
					name = LATr["Achievement sub category font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = subCatFontSetter,
					width = "full",
					order = 17,
				},
				subCategoryFontOutline = {
					name = LATr["Achievement sub category font outline"],
					type = "select",
					values = outlines,
					set = subCatFontSetter,
					width = "full",
					order = 18,
				},
				subCategoryCollapseButtonSize = {
					name = LATr["Sub category collapse Button size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 19,
				},
				achievementFont = {
					name = LATr["Achievement font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = achFontSetter,
					width = "full",
					order = 20,
				},
				achievementFontSize = {
					name = LATr["Achievement font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = achFontSetter,
					width = "full",
					order = 21,
				},
				achievementFontOutline = {
					name = LATr["Achievement font outline"],
					type = "select",
					values = outlines,
					set = achFontSetter,
					width = "full",
					order = 22,
				},
				achievementImageSize = {
					name = LATr["Achievement image size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 23,
				},
				achievementCriteriaFont = {
					name = LATr["Achievement criterias font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = achCritFont,
					width = "full",
					order = 24,
				},
				achievementCriteriaFontSize = { 
					name = LATr["Achievement criterias font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = achCritFont,
					width = "full",
					order = 25,
				},
				achievementCriteriaFontOutline = {
					name = LATr["Achievement criterias font outline"],
					type = "select",
					values = outlines,
					set = achCritFont,
					width = "full",
					order = 26,
				},
				numBooleanCirteriasColumns = {
					name = LATr["Number of columns in the achievement tracker for boolean criterias"],
					type = "range",
					min = 1,
					max = 8,
					step = 1,
					width = "full",
					order = 27,
				},
				maxWidth = {
					name = LATr["Achievement tracker column max width"],
					type="range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 28,
				},
				subCategoryCellPadding = {
					name = LATr["Sub category cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 29,
				},
				achievementCellPadding = {
					name = LATr["Achievement cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 30,
				},
				criteriasCellPadding = {
					name = LATr["Achievement criterias cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 31,
				},
				scale = {
					name = LATr["Achievement tracker scale"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 32,
				},
				backDropColor = {
					name = LATr["Achievement tracker background color"],
					type = "color",
					hasAlpha = true,
					get = colorGetter,
					set = colorSetter,
					width = "full",
					order = 33,
				},
				borderColor = {
					name = LATr["Achievement tracker border color"],
					type = "color",
					hasAlpha = true,
					get = colorGetter,
					set = colorSetter,
					width = "full",
					order = 34,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("achievementTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:GetModuleEnabled("achievementTracker"))
	core:RegisterModuleOptions("achievementTracker", getOptions, module.modName)
	chdb.achievementTracked = chdb.achievementTracked or {}
	chdb.categoryDisplay = chdb.categoryDisplay or {}
	self:SecureHook("AddTrackedAchievement")
	self:SecureHook("RemoveTrackedAchievement")
	self:SecureHook("GetNumTrackedAchievements")
	self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
end

function module:OnEnable()
	cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
	acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "updateAchievementsTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "hideAchievementsTracker")
	end
	for aid in pairs(chdb.achievementTracked) do
		AddTrackedAchievement(aid, true)
	end
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end

function module:OnDisable()
	anchor:Hide()
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
	acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "updateAchievementsTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "hideAchievementsTracker")
	end
end

---------------------------
--- Main Module Program ---
---------------------------
local function formatMoney(money)
	local gold = floor(money / 10000)
	local silver = mod(floor(money / 100), 100)
	local copper = mod(floor(money), 100)
	if gold > 0 then
		if silver > 0 then
			if copper > 0 then
				fMoney = format("|cff%s%d|r.|cff%s%d|r.|cff%s%d|r", COLOR_GOLD, gold, COLOR_SILVER, silver, COLOR_COPPER, copper)
			else
				fMoney = format("|cff%s%d|r.|cff%s%d|r", COLOR_GOLD, gold, COLOR_SILVER, silver)
			end
		elseif copper > 0 then
			fMoney = format("|cff%s%d|r.|cff%s%d|r", COLOR_GOLD, gold, COLOR_COPPER, copper)
		else
			fMoney = format("|cff%s%d|r", COLOR_GOLD, gold)
		end
	elseif silver > 0 then
		if copper > 0 then 
			fMoney = format("|cff%s%d|r.|cff%s%d|r", COLOR_SILVER, silver, COLOR_COPPER, copper)
		else
			fMoney = format("|cff%s%d|r", COLOR_SILVER, silver)
		end
	else
		fMoney = format("|cff%s%d|r", COLOR_COPPER, copper)
	end
	return fMoney
end

--------------
--- Events ---
--------------
function module:TRACKED_ACHIEVEMENT_UPDATE(event, aid)
	if achievements[aid] and achievements[aid].numCriterias > 0 then
		for cid = 1, achievements[aid].numCriterias do
			local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
			achievements[aid].criterias[cid] = new()
			achievements[aid].criterias[cid].name = cString
			achievements[aid].criterias[cid].cType = cType
			achievements[aid].criterias[cid].completed = completed or 0
			achievements[aid].criterias[cid].quantity = quantity
			achievements[aid].criterias[cid].reqQuantity = reqQuantity
		end
	end
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end

-------------
--- Hooks ---
-------------
function module:AddTrackedAchievement(aid, batch)
	numTrackedAchievement = numTrackedAchievement + 1
	chdb.achievementTracked[aid] = true
	achievements[aid] = new()
	_, achievements[aid].name, _, achievements[aid].completed, _, _, _, achievements[aid].description, _, achievements[aid].image = GetAchievementInfo(aid)
	achievements[aid].numCriterias = GetAchievementNumCriteria(aid)
	achievements[aid].criterias = new()
	for cid = 1, achievements[aid].numCriterias do
		local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
		achievements[aid].criterias[cid] = new()
		achievements[aid].criterias[cid].name = cString
		achievements[aid].criterias[cid].cType = cType
		achievements[aid].criterias[cid].completed = completed or 0
		achievements[aid].criterias[cid].quantity = quantity
		achievements[aid].criterias[cid].reqQuantity = reqQuantity
	end
	local categoryName, subCategoryName
	achievements[aid].subCategory = GetAchievementCategory(aid)
	subCategoryName, achievements[aid].category = GetCategoryInfo(achievements[aid].subCategory)
	categoryName = GetCategoryInfo(achievements[aid].category)
	if achievements[aid].category and achievements[aid].category > 0 then
		if not category[achievements[aid].category] then
			category[achievements[aid].category] = new()
			category[achievements[aid].category].subCategory = new()
			category[achievements[aid].category].numSubCategory = 0
			category[achievements[aid].category].achievements = new()
			category[achievements[aid].category].numAchievements = 0
			category[achievements[aid].category].name = categoryName
			category[achievements[aid].category].numTA, category[achievements[aid].category].numTCA  = GetCategoryNumAchievements(achievements[aid].category)
		end
		if not category[achievements[aid].category].subCategory[achievements[aid].subCategory] then
			category[achievements[aid].category].subCategory[achievements[aid].subCategory] = new()
			category[achievements[aid].category].numSubCategory = category[achievements[aid].category].numSubCategory + 1
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = 0
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements = new()
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].name = subCategoryName
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTA, category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(achievements[aid].subCategory)
		end
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements + 1
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements[aid] = true
	elseif achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		if not category[achievements[aid].subCategory] then
			category[achievements[aid].subCategory] = new()
			category[achievements[aid].subCategory].subCategory = new()
			category[achievements[aid].subCategory].numSubCategory = 0
			category[achievements[aid].subCategory].achievements = new()
			category[achievements[aid].subCategory].numAchievements = 0
			category[achievements[aid].subCategory].name = subCategoryName
			category[achievements[aid].subCategory].numTA, category[achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(achievements[aid].subCategory)
		end
		category[achievements[aid].subCategory].numAchievements = category[achievements[aid].subCategory].numAchievements + 1
		category[achievements[aid].subCategory].achievements[aid] = true
	end
	if not batch then
		LibQTip:Release(achievementTracker)
		achievementTracker = nil
		module:updateAchievementTracker()
	end
end

function module:RemoveTrackedAchievement(aid)
	numTrackedAchievement = numTrackedAchievement - 1
	chdb.achievementTracked[aid] = nil
	if achievements[aid].category and achievements[aid].category > 0 and achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements[aid] = nil
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements - 1
		if category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements == 0 then
			category[achievements[aid].category].subCategory[achievements[aid].subCategory] = del(category[achievements[aid].category].subCategory[achievements[aid].subCategory])
			category[achievements[aid].category].numSubCategory = category[achievements[aid].category].numSubCategory - 1
			if category[achievements[aid].category].numSubCategory == 0 and category[achievements[aid].category].numAchievements == 0 then
				category[achievements[aid].category] = del(category[achievements[aid].category])
			end
		end
	elseif achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		category[achievements[aid].subCategory].achievements[aid] = nil
		category[achievements[aid].subCategory].numAchievements = category[achievements[aid].subCategory].numAchievements - 1
		if category[achievements[aid].subCategory].numSubCategory == 0 and category[achievements[aid].subCategory].numAchievements == 0 then
			category[achievements[aid].subCategory] = del(category[achievements[aid].subCategory])
		end
	end
	achievements[aid] = del(achievements[aid])
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end

function module:GetNumTrackedAchievements()
	return numTrackedAchievement
end

function module:updateAchievementsTracker(event)
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end

function module:hideAchievementsTracker(event)
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if bType == "category" then
			if chdb.categoryDisplay[bValue] == 1 then
				chdb.categoryDisplay[bValue] = 0
			else
				chdb.categoryDisplay[bValue] = 1
			end
			achievementTracker:SetScript("OnLeave", nil)
			LibQTip:Release(achievementTracker)
			achievementTracker = nil
			module:updateAchievementTracker()
		elseif bType == "achievement" then
			if button == "RightButton" then
				if IsShiftKeyDown() then
					ChatFrameEditBox:Show()
					ChatEdit_InsertLink(GetAchievementLink(bValue))
				else
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetAchievementLink(bValue)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsShiftKeyDown() then
					RemoveTrackedAchievement(tonumber(bValue))
				else
					if AchievementFrame:IsShown() then
						HideUIPanel(AchievementFrame);
					else
						ShowUIPanel(AchievementFrame);
					end
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if bType == "achievement" then
			GameTooltip:SetOwner(tooltip, anchor)
			GameTooltip:SetHyperlink(GetAchievementLink(bValue))
		end
	end)
	self:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateAchievementTracker()
	if (not prdb.toggleTracker) or (numTrackedAchievement == 0) then return end
	achievementTracker = LibQTip:Acquire("tomQuest2Achievements", prdb.numBooleanCirteriasColumns + 1)
	achievementTracker:EnableMouse(true)
	achievementTracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
	achievementTracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
	achievementTracker:SetScale(prdb.scale)
	achievementTracker:SetFrameStrata("BACKGROUND")
	local tempLine
	for cid, cData in pairs(category) do
		chdb.categoryDisplay[cid] = chdb.categoryDisplay[cid] or 1
		tempLine = achievementTracker:AddLine()
 		achievementTracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.categoryDisplay[cid] == 0 and "Plus" or "Minus", prdb.categoryCollapseButtonSize, core:getColorFromCompletionHEX(cData.numTCA / cData.numTA) or "ffffff", cData.name), cATrFont, "LEFT", prdb.numBooleanCirteriasColumns, cProvider, "category", cid, prdb.maxWidth, 0)
 		achievementTracker:SetCell(tempLine, prdb.numBooleanCirteriasColumns + 1, format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(cData.numTCA / cData.numTA), cData.numTCA , cData.numTA), cATrFont, "RIGHT", 1, cProvider, "category", cid, prdb.maxWidth, 0)
		if chdb.categoryDisplay[cid] == 1 then
			if cData.numAchievements > 0 then
				for aid in pairs(cData.achievements) do
					tempLine = achievementTracker:AddLine()
 					achievementTracker:SetCell(tempLine, 1, format("|T%s:%d|t |cffffffff%s|r", achievements[aid].image, prdb.achievementImageSize, achievements[aid].name), aATrFont, "LEFT", prdb.numBooleanCirteriasColumns + 1, cProvider, "achievement", aid, prdb.maxWidth, prdb.subCategoryCellPadding, trackerAnchor)
					if achievements[aid].numCriterias and achievements[aid].numCriterias > 0 then
						local cIndex = 1
						for cid, acData in pairs (achievements[aid].criterias) do
							if (prdb.hideCompletedCriteria and achievements[aid].criterias[cid].completed == 0) or (not prdb.hideCompletedCriteria) then
								if acData.reqQuantity > 1 then
									tempLine = achievementTracker:AddLine()
									achievementTracker:SetCell(tempLine, 1, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity / acData.reqQuantity), acData.name), acATrFont, "LEFT", prdb.numBooleanCirteriasColumns, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
									if achievements[aid].criterias[cid].cType == 62 or achievements[aid].criterias[cid].cType == 67 then
										achievementTracker:SetCell(tempLine, prdb.numBooleanCirteriasColumns + 1, formatMoney(acData.quantity), acATrFont, "RIGHT", 1, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
									else
										achievementTracker:SetCell(tempLine, prdb.numBooleanCirteriasColumns + 1, format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(acData.quantity / acData.reqQuantity), acData.quantity, acData.reqQuantity), acATrFont, "RIGHT", 1, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
									end
								else
									if cIndex == 1 then
										tempLine = achievementTracker:AddLine()
									end
									if cIndex == prdb.numBooleanCirteriasColumns then
										achievementTracker:SetCell(tempLine, cIndex, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity), acData.name), acATrFont, "LEFT", 2, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
										cIndex = 1
									else
										achievementTracker:SetCell(tempLine, cIndex, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity), acData.name), acATrFont, "LEFT", 1, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
										cIndex = cIndex + 1
									end
								end
							end
						end
					else
						tempLine = achievementTracker:AddLine()
						achievementTracker:SetCell(tempLine, 1, achievements[aid].description, acATrFont, "LEFT", prdb.numBooleanCirteriasColumns + 1, cProvider, "achievement", aid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
					end
				end
			end
			if category[cid].numSubCategory > 0 then
				for scid, scData in pairs(cData.subCategory) do
					chdb.categoryDisplay[scid] = chdb.categoryDisplay[scid] or 1
					tempLine = achievementTracker:AddLine()
 					achievementTracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.categoryDisplay[scid] == 0 and "Plus" or "Minus", prdb.subCategoryCollapseButtonSize, core:getColorFromCompletionHEX(scData.numTCA / scData.numTA), scData.name), scATrFont, "LEFT", prdb.numBooleanCirteriasColumns, cProvider, "category", scid, prdb.maxWidth, prdb.subCategoryCellPadding)
 					achievementTracker:SetCell(tempLine, prdb.numBooleanCirteriasColumns + 1, format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(scData.numTCA / scData.numTA), scData.numTCA , scData.numTA), scATrFont, "RIGHT", 1, cProvider, "category", scid, prdb.maxWidth, prdb.subCategoryCellPadding)
					if chdb.categoryDisplay[scid] == 1 then
						if scData.numAchievements > 0 then
							for aid in pairs(scData.achievements) do
								tempLine = achievementTracker:AddLine()
 								achievementTracker:SetCell(tempLine, 1, format("|T%s:%d|t |cffffffff%s|r", achievements[aid].image, prdb.achievementImageSize, achievements[aid].name), aATrFont, "LEFT", prdb.numBooleanCirteriasColumns + 1, cProvider, "achievement", aid, prdb.maxWidth, prdb.achievementCellPadding, trackerAnchor)
								if achievements[aid].numCriterias and achievements[aid].numCriterias > 0 then
									local cIndex = 1
									for cid, acData in pairs (achievements[aid].criterias) do
										if (prdb.hideCompletedCriteria and achievements[aid].criterias[cid].completed == 0) or (not prdb.hideCompletedCriteria) then
											if acData.reqQuantity > 1 then
												tempLine = achievementTracker:AddLine()
												achievementTracker:SetCell(tempLine, 1, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity / acData.reqQuantity), acData.name), acATrFont, "LEFT", prdb.numBooleanCirteriasColumns, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
												if achievements[aid].criterias[cid].cType == 62 or achievements[aid].criterias[cid].cType == 67 then
													achievementTracker:SetCell(tempLine, prdb.numBooleanCirteriasColumns + 1, formatMoney(acData.quantity), acATrFont, "RIGHT", 1, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
												else
													achievementTracker:SetCell(tempLine, prdb.numBooleanCirteriasColumns + 1, format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(acData.quantity / acData.reqQuantity), acData.quantity, acData.reqQuantity), acATrFont, "RIGHT", 1, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
												end
											else
												if cIndex == 1 then
													tempLine = achievementTracker:AddLine()
												end
												if cIndex == prdb.numBooleanCirteriasColumns then
													achievementTracker:SetCell(tempLine, cIndex, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity), acData.name), acATrFont, "LEFT", 2, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
													cIndex = 1
												else
													achievementTracker:SetCell(tempLine, cIndex, format("|cff%s%s|r", core:getColorFromCompletionHEX(acData.quantity), acData.name), acATrFont, "LEFT", 1, cProvider, "criterias", cid, prdb.maxWidth, prdb.criteriasCellPadding, trackerAnchor)
													cIndex = cIndex + 1
												end
											end
										end
									end
								end
							end
						end
					end
				end
			end
		end
	end
	
	achievementTracker:SetClampedToScreen(true)
	achievementTracker:ClearAllPoints()
	if prdb.anchorPosition == 1 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT")
		else
			achievementTracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, 24)
		end
	elseif prdb.anchorPosition == 2 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("TOP", anchor, "BOTTOM")
		else
			achievementTracker:SetPoint("TOP", anchor, "BOTTOM", 0, 24)
		end
	elseif prdb.anchorPosition == 3 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT")
		else
			achievementTracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT", 0, 24)
		end
	elseif prdb.anchorPosition == 4 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT")
		else
			achievementTracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -24)
		end
	elseif prdb.anchorPosition == 5 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("BOTTOM", anchor, "TOP")
		else
			achievementTracker:SetPoint("BOTTOM", anchor, "TOP", 0, -24)
		end
	elseif prdb.anchorPosition == 6 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT")
		else
			achievementTracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT", 0, -24)
		end
	end
	achievementTracker:Show()
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2ACHIEVEMENTTRACKERTOGGLE = LATr["Achievement Tracker"]

function tQ2toggleAchievementTracker()
	prdb.toggleTracker = not prdb.toggleTracker
	LibQTip:Release(achievementTracker)
	module:updateAchievementTracker()
end